#!/bin/bash
# bangban: hangmanもどき by Masahiro Nakajma (2017/1/16) for Software Design

usage_exit() {
  if (($1)) ; then echo "$0: No dictionary found."; fi
  echo "Usage: $0 [-m minimum_word_length] [-r maximum_number_of_failures] [-d dictionary]" 1>&2
  exit 1
}

game_exit() {
  [ -f $tempdict ] && rm $tempdict
  reset
  print_score
  exit 0
}

print_status()
{
  echo ; echo -n "    Word : " 
  for c in ${ans[@]}; do echo -n $c; done ; echo
  echo -n " Guessed : "
  for c in ${guessed[@]}
  do
    echo $c
  done | sort | tr -cd [:print:] 
  # UTFの絵文字を使っている。表示できない環境では、別の文字に置き換える。
  # Skull: \U1F480, Fire: \U1F525, Spark: \U1F4A5, Bomb: \U1F4A3
  local skull="\xf0\x9f\x92\x80" fire="\xf0\x9f\x94\xa5" spark="\xf0\x9f\x92\xa5" bomb="\xf0\x9f\x92\xa3"
  echo -n -e "\n    ${skull} "; printf "(%2d/%2d):" ${failure} ${mfailures}
  for ((i=0;i<failure-1;i++)) ; do echo -n " "; done
  if ((failure>0)) ;then echo -n -e "${fire}" ; fi
  for ((i=0;i<mfailures-failure-1;i++)) ; do echo -n "_"; done
  if ((failure==mfailures)) ;then echo -e "${spark}${fire}" ;else echo -e "${bomb}" ; fi
}

print_score()
{
  echo "You got ${got} word(s) with ${challenge} challenge(s)."
  echo
}


not_in_list()
{ # リスト中に文字が「無いこと」を確認
  local c=$1
  shift
  local list=($@)
  for cc in ${list[@]}
  do
    if [ $cc = $c ]
    then
        return 1
    fi
  done
  return 0
}

not_in_word()
{ # 単語中に文字が「無いこと」を確認 & ある場合はansに記入
  local c=$1
  local r=0
  for ((i=0; i<${#w[@]}; i++))
  do
    if [ ${w[$i]} = $c ]
    then
        r=1
        ans[$i]=$c
    fi
  done
  return $r
}

dictionary="/usr/share/dict/words" # bash on windowsには入っていない
length=6 mfailures=7 challenge=0 got=0

while getopts m:d:r: OPT
do
  case $OPT in
    m)  length=$OPTARG ;;
    r)  mfailures=$OPTARG ;;
    d)  dictionary=$OPTARG ;;
    \?) usage_exit ;;
  esac
done

[ -f $dictionary ] || usage_exit 1
if (( $length > 11 )) ; then length=11; fi
if (( $mfailures < 1 )) ; then mfailures=7; fi

echo "minimum word length : ${length}"
echo "maximum number of failures : ${mfailures}"
echo "dictionary : ${dictionary}"

trap 'game_exit' 1 2 15

tempdict=$(mktemp /tmp/sdbangban.XXXXXX)
grep -v \' ${dictionary} | tr A-Z a-z | sort | uniq > $tempdict
dl=$(wc -l ${tempdict}) ; dl=${dl% *} ; dl=${dl#* } # 単語数を取得

tic=("|" "/" "-" "\\") # 「くるくる」演出用
playing=1
while ((playing))
do
  word=""
  echo
  while  $(test "$word" = "")
  do
    echo -n "."
    # ∵) 0 <= RANDOM <= 32767 なので、Macの辞書は234371ほど、Ubuntuで71972で足りない
    word=$(sed -n "$((RANDOM * RANDOM % dl + 1))p" ${tempdict})
    if (( ${#word} < length )) ; then word=""; fi
  done

  w=()
  ans=()
  while read -n 1 c
  do
    if [ "$c" != "" ] ; then w+=("$c") ans+=("-"); fi  # read -n 1 で最後に生じる空文字を除外するための処理
  done <<< $word

  echo 
  guessed=()
  failure=0
  challenge=$((challenge+1))
  
  print_status
  while :
  do
    while read -s -n 1 -p "Guess:" c
    do 
      case $c in
        [A-Z]) c=$(tr A-Z a-z <<< ${c}); break ;; # 大文字を小文字に変換 bash4では ${c,} でOK
        [a-z]) break ;;
        *) printf "\nNot a valid guess: '%c'\n" ${c}; continue ;;
      esac
    done
    echo $c
    if not_in_list "$c" ${guessed[@]} # not in guessed
    then
      guessed+=("$c")
      if not_in_word "$c" # not in word
      then 
        failure=$((failure+1))
      fi
    else
      echo "Already guessed '$c'"
    fi

    print_status
    if not_in_list "-" ${ans[@]}
    then
      echo
      echo "You got it! Yes, the word was \"${word}\""
      got=$((got+1))
      break
    elif (( failure >= mfailures ))
    then
      echo
      echo "Sorry, the word was \"${word}\""
      break
    fi
  done
  print_score

  # リトライ前の「くるくる」演出 : 無用な場合はコメントアウト
  echo -n -e "\033[?25l"  # カーソルを表示しないエスケープシーケンス
  for ((i=0; i < ${#tic[@]} * 5 + 1; i++)) ; do m=$((i%${#tic[@]})); sleep 0.1; printf "%c\r" ${tic[$m]}; done ; # sleepに小数点を使っているため、動かない環境もあるので注意
  echo -n -e "\033[?25h" # カーソルを表示するエスケープシーケンス
  
  while read -s -n 1 -p "Another word? " retry
  do
    echo $retry
    case $retry in
      [Yy]) break ;;
      [Nn]) playing=0; break ;;
      *) echo "Please type 'y' or 'n'" ;;
    esac
  done
done

game_exit
